from pydantic import BaseModel, Field
from typing import Optional

from quartz_solar_forecast.inverters.enphase import EnphaseInverter, EnphaseSettings
from quartz_solar_forecast.inverters.givenergy import GivEnergySettings, GivEnergyInverter
from quartz_solar_forecast.inverters.mock import MockInverter
from quartz_solar_forecast.inverters.solarman import SolarmanSettings, SolarmanInverter
from quartz_solar_forecast.inverters.solis import SolisSettings, SolisInverter
from quartz_solar_forecast.inverters.victron import VictronSettings, VictronInverter


class PVSite(BaseModel):
    latitude: float = Field(..., description="the latitude of the site", ge=-90, le=90)
    longitude: float = Field(
        ..., description="the longitude of the site", ge=-180, le=180
    )
    capacity_kwp: float = Field(..., description="the capacity [kwp] of the site", ge=0)
    tilt: float = Field(
        default=35,
        description="the tilt of the site [degrees], the panels' angle relative to horizontal ground",
        ge=0,
        le=90,
    )
    orientation: float = Field(
        default=180,
        description="the orientation of the site [degrees], the angle between north and the direction the panels face, measured on the horizontal plane.",
        ge=0,
        le=360,
    )
    inverter_type: str = Field(
        default=None,
        description="The type of inverter used",
        json_schema_extra=["enphase", "solis", "givenergy", "solarman", None],
    )

    def round_latitude_and_longitude(self):
        """ Round the latitude and longitude to 2 decimal places

        This is to ensure that the location of the site is not stored exactly.
        """
        self.latitude = round(self.latitude, 2)
        self.longitude = round(self.longitude, 2)

    def get_inverter(self):
        if self.inverter_type == 'enphase':
            return EnphaseInverter(EnphaseSettings())
        elif self.inverter_type == 'solis':
            return SolisInverter(SolisSettings())
        elif self.inverter_type == 'givenergy':
            return GivEnergyInverter(GivEnergySettings())
        elif self.inverter_type == 'solarman':
            return SolarmanInverter(SolarmanSettings())
        elif self.inverter_type == 'victron':
            return VictronInverter.from_settings(VictronSettings())
        else:
            return MockInverter()

class ForecastRequest(BaseModel):
    site: PVSite
    timestamp: Optional[str] = None

class TokenRequest(BaseModel):
    redirect_url: str


from typing import Callable

import pandas as pd
from datetime import datetime, timedelta
from quartz_solar_forecast.inverters.inverter import AbstractInverter
from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict
try:
    from ocf_vrmapi.vrm import VRM_API
except:
    print("Tried to import ocf_vrmapi but couldnt, Victron Inverter functions won't work")

class VictronSettings(BaseSettings):
    model_config = SettingsConfigDict(env_file='.env', extra='ignore')

    username: str = Field(alias="VICTRON_USER")
    password: str = Field(alias="VICTRON_PASS")


class VictronInverter(AbstractInverter):

    def __init__(self, get_sites: Callable, get_kwh_stats: Callable):
        self.__get_sites = get_sites
        self.__get_kwh_stats = get_kwh_stats

    @classmethod
    def from_settings(cls, settings: VictronSettings):
        api = VRM_API(username=settings.username, password=settings.password)
        get_sites = lambda: api.get_user_sites(api.user_id)
        end = datetime.now()
        start = end - timedelta(weeks=1)
        get_kwh_stats = lambda site_id: api.get_kwh_stats(site_id, start=start, end=end)
        return cls(get_sites, get_kwh_stats)

    def get_data(self, ts: pd.Timestamp) -> pd.DataFrame:
        sites = self.__get_sites()
        # get first site (bit of a guess)
        first_site_id = sites["records"][0]["idSite"]

        stats = self.__get_kwh_stats(first_site_id)

        kwh = stats["records"]["kwh"]

        df = pd.DataFrame(kwh)

        df[0] = pd.to_datetime(df[0], unit='ms')
        df.columns = ["timestamp", "power_kw"]
        return df



import os
import pandas as pd
from datetime import datetime, timedelta
from quartz_solar_forecast.forecast import run_forecast
from quartz_solar_forecast.pydantic_models import PVSite


def generate_all_forecasts(
    init_time_freq: int,
    start: datetime,
    end: datetime,
    latitude: float,
    longitude: float,
    capacity_kwp: float,
) -> pd.DataFrame:

    all_forecasts = pd.DataFrame()

    init_time = start
    while init_time <= end:
        print(f"Running forecast for initialization time: {init_time}")
        predictions_df = forecast_for_site(
            latitude, longitude, capacity_kwp, init_time=init_time
        )
        predictions_df["forecast_init_time"] = init_time
        all_forecasts = pd.concat([all_forecasts, predictions_df])
        init_time += timedelta(hours=init_time_freq)

    return all_forecasts


def forecast_for_site(
    latitude: float,
    longitude: float,
    capacity_kwp: float,
    model: str = "gb",
    init_time: datetime = None,
) -> pd.DataFrame:

    site = PVSite(latitude=latitude, longitude=longitude, capacity_kwp=capacity_kwp)
    predictions_df = run_forecast(site=site, model=model, ts=init_time)
    predictions_df.reset_index(inplace=True)
    predictions_df.rename(columns={"index": "datetime"}, inplace=True)
    return predictions_df


def write_out_forecasts(
    init_time_freq,
    start_datetime,
    end_datetime,
    site_name,
    latitude,
    longitude,
    capacity_kwp,
):
    """
    Generates forecasts at specified intervals and saves them into a CSV file.

    Args:
        init_time_freq (int): The frequency in hours at which the forecasts are generated.
        start_datetime (str): The starting date and time for generating forecasts.
        end_datetime (str): The ending date and time for generating forecasts.
        site_name (str): The name of the site for which the forecasts are generated.
        latitude (float): The latitude of the PV site.
        longitude (float): The longitude of the PV site.
        capacity_kwp (float): The capacity of the PV site in kilowatts peak (kWp).
    """
    start = datetime.strptime(start_datetime, "%Y-%m-%d %H:%M:%S")
    start_date = start.date()
    end = datetime.strptime(end_datetime, "%Y-%m-%d %H:%M:%S")
    end_date = end.date()
    all_forecasts = generate_all_forecasts(
        init_time_freq, start, end, latitude, longitude, capacity_kwp
    )

    output_dir = os.path.join(os.getcwd(), "csv_forecasts")
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    output_file_name = f"forecast_{site_name}_{start_date}_{end_date}.csv"
    output_file_path = os.path.join(output_dir, output_file_name)
    all_forecasts.to_csv(output_file_path, index=False)
    print(f"Forecasts saved to {output_file_path}")




""" Make script to run the evaluation on the test set.

The idea is to run the model on the test set and then compare the results to the actual PV generation.
The NWP (ICON) and PV data are both pulled Open Climate Fix's Hugging Face page.

Please note it can take hours to pull the NWP data.
The data will be cached locally so next time you run it, itll be much quicker
"""

from quartz_solar_forecast.evaluation import run_eval

if __name__ == '__main__':
    run_eval()



""" Example code to run the forecast"""
from quartz_solar_forecast.forecast import run_forecast
from quartz_solar_forecast.pydantic_models import PVSite
from datetime import datetime, timedelta

def main():
    # make input data
    site = PVSite(latitude=51.75, longitude=-1.25, capacity_kwp=1.25)
    
    ts = datetime.today() - timedelta(weeks=1)

    # User has three options for the 'nwp_source': 'icon', 'gfs', or 'ukmo'.
    predictions_df = run_forecast(site=site, ts=ts, nwp_source="icon")

    print(predictions_df)
    print(f"Max: {predictions_df['power_kw'].max()}")

if __name__ == "__main__":
    main()



import os

import pytest
from fastapi.testclient import TestClient

from api.app.api import app

expected_prediction_key = "power_kw"
expected_dict_keys = ["timestamp", "predictions"]
prediction_key = expected_dict_keys[1]
expected_number_of_values = 192
expected_response_on_wrong_types = {
    "detail": [
        {
            "loc": ["body", "site", "latitude"],
            "msg": "Input should be less than or equal to 90",
            "type": "less_than_equal",
            "input": 91,
            "ctx": {"le": 90.0},
        },
        {
            "loc": ["body", "site", "longitude"],
            "msg": "Input should be greater than or equal to -180",
            "type": "greater_than_equal",
            "input": -181,
            "ctx": {"ge": -180.0},
        },
        {
            "loc": ["body", "site", "capacity_kwp"],
            "msg": "Input should be greater than or equal to 0",
            "type": "greater_than_equal",
            "input": -1,
            "ctx": {"ge": 0.0},
        },
        {
            "loc": ["body", "site", "tilt"],
            "msg": "Input should be less than or equal to 90",
            "type": "less_than_equal",
            "input": 91,
            "ctx": {"le": 90.0},
        },
        {
            "loc": ["body", "site", "orientation"],
            "msg": "Input should be less than or equal to 360",
            "type": "less_than_equal",
            "input": 361,
            "ctx": {"le": 360.0},
        },
    ]
}

envs = {
    "ENPHASE_CLIENT_ID": "1",
    "ENPHASE_SYSTEM_ID": "1",
    "ENPHASE_API_KEY": "test_key",
    "ENPHASE_CLIENT_SECRET": "secret_secret",
}


@pytest.fixture
def client():
    return TestClient(app)


@pytest.fixture
def body():
    return {
        "site": {
            "latitude": -90,
            "longitude": -180,
            "capacity_kwp": 0,
            "tilt": 35,
            "orientation": 180,
            "inverter_type": "string",
        },
        "timestamp": "2021-01-26 01:15:00",
    }


@pytest.fixture
def body_short():
    return {
        "site": {
            "latitude": -90,
            "longitude": -180,
            "capacity_kwp": 0,
        },
        "timestamp": "2021-01-26 01:15:00",
    }


@pytest.fixture
def body_wrong():
    return {
        "site": {
            "latitude": 91,
            "longitude": -181,
            "capacity_kwp": -1,
            "tilt": 91,
            "orientation": 361,
            "inverter_type": "string",
        },
        "timestamp": "2021-01-26 01:15:00",
    }


@pytest.fixture
def bad_redirect_url(client):
    return "url"


@pytest.fixture
def bad_redirect_url_correct_param(client):
    return "url?code=code"


def test_api_ok(client, body):
    response = client.post("/forecast/", json=body)
    response_body = response.json()
    assert response.status_code == 200
    assert isinstance(response_body[prediction_key][expected_prediction_key], dict)
    assert all(
        [key in response_body.keys() for key in expected_dict_keys]
    ), "Expected dictonary key is missing"
    assert (
        len(response_body[prediction_key][expected_prediction_key]) == expected_number_of_values
    ), "Expected number of values is wrong"


def test_api_ok_short(client, body_short):
    response = client.post("/forecast/", json=body_short)
    response_body = response.json()

    assert response.status_code == 200
    assert isinstance(response_body[prediction_key][expected_prediction_key], dict)
    assert all(
        [key in response_body.keys() for key in expected_dict_keys]
    ), "Expected dictonary key is missing"
    assert len(response_body[prediction_key][expected_prediction_key]) == expected_number_of_values


def test_api_wrong_body(client, body_wrong):
    # Tests all wrong data types responses
    response = client.post("/forecast/", json=body_wrong)
    response_body = response.json()

    assert response.status_code == 422
    assert response_body == expected_response_on_wrong_types


def test_getenphse_authorization_url(client, monkeypatch):

    monkeypatch.setattr(os, "environ", envs)

    response = client.get("/solar_inverters/enphase/auth_url")
    assert response.status_code == 200
    assert isinstance(response.json()["auth_url"], str)


def test_getenphse_token_and_system_id_bad_redirect_url(client, bad_redirect_url, monkeypatch):

    monkeypatch.setattr(os, "environ", envs)

    response = client.post(
        "/solar_inverters/enphase/token_and_id", json={"redirect_url": bad_redirect_url}
    )
    response_body = response.json()
    assert response.status_code == 400
    assert response_body["detail"] == "Invalid redirect URL"


def test_getenphse_token_and_system_id_bad_redirect_url(
    client, bad_redirect_url_correct_param, monkeypatch
):

    monkeypatch.setattr(os, "environ", envs)

    response = client.post(
        "/solar_inverters/enphase/token_and_id",
        json={"redirect_url": bad_redirect_url_correct_param},
    )
    response_body = response.json()
    assert response.status_code == 400
    assert "Error getting access token and system ID: " in response_body["detail"]



from typing import Optional

import requests
import pandas as pd
from datetime import timedelta, datetime
from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

from quartz_solar_forecast.inverters.inverter import AbstractInverter


class SolarmanSettings(BaseSettings):
    model_config = SettingsConfigDict(env_file='.env', extra='ignore')

    url: str = Field(alias="SOLARMAN_API_URL")
    token: str = Field(alias="SOLARMAN_TOKEN")
    id: str = Field(alias="SOLARMAN_ID")


class SolarmanInverter(AbstractInverter):

    def __init__(self, settings: SolarmanSettings):
        self.__settings = settings

    def get_data(self, ts: pd.Timestamp) -> Optional[pd.DataFrame]:
        try:
            end_date = datetime.now()
            start_date = end_date - timedelta(weeks=1)
            solarman_data = get_solarman_data(start_date, end_date, self.__settings)

            # Filter out rows with null power_kw values
            valid_data = solarman_data.dropna(subset=['power_kw'])

            if valid_data.empty:
                print("No valid Solarman data found.")
                return pd.DataFrame(columns=['timestamp', 'power_kw'])

            return valid_data
        except Exception as e:
            print(f"Error retrieving Solarman data: {str(e)}")
            return pd.DataFrame(columns=['timestamp', 'power_kw'])


def get_solarman_data(start_date, end_date, settings: SolarmanSettings):
    """
    Fetch data from the Solarman API from start_date to end_date.
    
    :param start_date: Start date (datetime object)
    :param end_date: End date (datetime object)
    :param settings: the Solarman settings
    :return: DataFrame with timestamp and power_kw columns
    """
    all_data = []
    
    current_date = start_date
    
    while current_date <= end_date:
        year = current_date.year
        month = current_date.month
        day = current_date.day
        
        url = f"{settings.url}/{settings.id}/record"
        
        headers = {
            'Authorization': f'Bearer {settings.token}',
            'User-Agent': 'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.9.0.7) Gecko/2009021910 Firefox/3.0.7'
        }
        
        params = {
            'year': year,
            'month': month,
            'day': day
        }
        
        response = requests.get(url, headers=headers, params=params)
        
        if response.status_code != 200:
            print(f"API request failed for {current_date} with status code {response.status_code}")
        else:
            data = response.json()
            records = data.get('records', [])
            
            if records:
                df = pd.DataFrame(records)
                df['timestamp'] = pd.to_datetime(df['dateTime'], unit='s')
                all_data.append(df)
        
        current_date += timedelta(days=1)
    
    if not all_data:
        raise ValueError("No data found for the specified date range")
    
    # Concatenate all dataframes
    full_data = pd.concat(all_data, ignore_index=True)
    
    # Select relevant columns
    full_data = full_data[['timestamp', 'generationPower']]

    # Convert watts to kilowatts and rename the column
    full_data['power_kw'] = full_data['generationPower'] / 1000.0
    full_data = full_data.drop('generationPower', axis=1)

    # Ensure the dataframe only has 'timestamp' and 'power_kw' columns
    full_data = full_data[['timestamp', 'power_kw']]

    # Sort by timestamp
    full_data = full_data.sort_values('timestamp')
    
    return full_data



from quartz_solar_forecast.eval.metrics import metrics
import pandas as pd
import numpy as np


def test_metrics():

    # create a fake dataframe

    results_df = pd.DataFrame(
        columns=[
            "pv_id",
            "timestamp",
            "horizon_hour",
            "forecast_power",
            "generation_power",
        ],
        data=np.random.random((100, 5)),
    )

    pv_metadata = pd.DataFrame(
        columns=[
            "pv_id",
            "capacity",
        ],
        data=np.random.random((100, 2)),
    )

    # call the metrics function
    metrics(results_df, pv_metadata)



""" Get nwp data from HF"""
import os
import sys
import pandas as pd
import numpy as np

import ocf_blosc2  # noqa
import xarray as xr

import multiprocessing

from quartz_solar_forecast.eval.utils import make_hf_filename

multiprocessing.set_start_method("spawn", force=True)


def get_nwp(time_locations: pd.DataFrame):
    """
    Get all the nwp data fpr the time locations

    time_locations should have the following columns:
    - timestamp
    - latitude
    - longitude
    - pv_id
    """

    all_nwp_dfs = []

    tasks_args = []
    with multiprocessing.Pool() as pool:
        for i, row in time_locations.iterrows():
            print(f"Making task {i} of {len(time_locations)}")

            kwargs = {
                "timestamp": row["timestamp"],
                "latitude": row["latitude"],
                "longitude": row["longitude"],
                "pv_id": row["pv_id"],
                "progress": np.round(i / len(time_locations), 3),
            }

            # collect together args for pool.starmap
            task_arg = list(kwargs.values())
            tasks_args.append(task_arg)

        print("Made all NWP tasks, now getting the data")
        results = pool.starmap(get_nwp_for_one_timestamp_one_location, tasks_args)

    print("Got all NWP data")

    for result in results:
        one_nwp_df = result

        all_nwp_dfs.append(one_nwp_df)

    all_nwp_df = pd.concat(all_nwp_dfs)

    return all_nwp_df


def get_nwp_for_one_timestamp_one_location(
    timestamp: pd.Timestamp, latitude, longitude, pv_id: int = None, progress: float = True
):
    """
    Get NWP data from Hugging Face for one timestamp and one location

    :param timestamp: the timestamp for when you want the forecast for
    :param latitude: the latitude of the location
    :param longitude: the longitude of the location
    :param pv_id: the pv_id of the location, if known
    :param progress: Float of how far through the process we are. This is becasue we use multiprocessing
        to pull lots of NWP data. This should be a float between 0 and 1


    :return: nwp forecast in xarray
    """

    # List which files are available. Not all dates, and model run times are available
    # fs = HfFileSystem()
    # print(fs.ls("datasets/openclimatefix/dwd-icon-eu/data/2022/4/11/", detail=False))

    # Ensure timestamp is a pd.Timestamp object
    if not isinstance(timestamp, pd.Timestamp):
        timestamp = pd.to_datetime(timestamp)

    # round timestamp to 6 hours floor
    timestamp_floor = timestamp.floor("6h")
    date_and_hour, huggingface_file = make_hf_filename(timestamp_floor)

    # dataset variables, note these are unique for ICON
    variables = ["t_2m", "tot_prec", "clch", "clcm", "clcl", "u", "v", "aswdir_s", "aswdifd_s"]

    cache_dir = "data/nwp"
    cache_file = f"{cache_dir}/{date_and_hour}_lat={latitude}_lon={longitude}.zarr"
    if not os.path.exists(cache_file):
        # use fsspec to copy file
        print(f"Copying file {huggingface_file} from HF to local")
        sys.stdout.flush()

        data = xr.open_zarr(
            f"{huggingface_file}",
            chunks="auto",
        )

        # take nearest location and only select the variables we want
        data_at_location = data.sel(latitude=latitude, longitude=longitude, method="nearest")
        data_at_location = data_at_location[variables]

        # choice the first isobaricInhPa
        data_at_location = data_at_location.isel(isobaricInhPa=-1)

        #  reduce to 54 hours timestamps, this means there is at least a 48 hours forecast
        data_at_location = data_at_location.isel(step=slice(0, 54))

        # load all the data, this can take about ~1 minute seconds
        print(f"Loading dataset for {timestamp=} {longitude=} {latitude=}")
        data_at_location.load()

        # save to cache
        print(f"Saving to cache {cache_file}")
        data_at_location.to_zarr(cache_file)
    else:
        # load from cache
        print("loading from cache")
        data_at_location = xr.open_zarr(cache_file)

    # make times from the init time + steps
    times = pd.to_datetime(data_at_location.time.values) + pd.to_timedelta(
        data_at_location.step.values, unit="h"
    )

    # convert to pandas dataframe
    df = pd.DataFrame(times, columns=["time"])
    for variable in variables:
        df[variable] = data_at_location[variable].values

    # make wind speed out of u and v
    df["si10"] = (df["u"] ** 2 + df["v"] ** 2) ** 0.5

    # rename variables
    df = df.rename(
        columns={
            "t_2m": "t",
            "tot_prec": "prate",
            "aswdifd_s": "dswrf",
            "aswdir_s": "dlwrf",
            "clcl": "lcc",
            "clcm": "mcc",
            "clch": "hcc",
        }
    )

    # add visbility for the moment
    # TODO
    df["vis"] = 10000

    # drop u and v
    df = df.drop(columns=["u", "v"])

    # rename id to pv_id
    df = df.rename(columns={"id": "pv_id"})

    # add columns for timestamp, latitude and longitude
    df["timestamp"] = timestamp
    df["latitude"] = latitude
    df["longitude"] = longitude

    # add pv_id columns if it is given
    if pv_id is not None:
        df["pv_id"] = pv_id

    if progress:
        print(f"Getting NWP for {timestamp} {pv_id}. Progress: {100*progress}%")

    return df
